#ifndef __OBJDIFF_H
#define __OBJDIFF_H

#include <set>
#include <map>
#include <string>
#include <math.h>

#define MSE_POW 8

#include "imageprocess.h"
#include "rawimage.h"

#include "codecregistry.h"


using namespace std;


/**
 * Supported compare modes
 */
typedef enum  {
	/** Comparing regular objects locally */
	OBJ_DIFF_MODE_NORMAL,
	/** Comparing a local object to data collected from a remote server */
	OBJ_DIFF_MODE_REMOTE,
	/** Comparing huge objects locally */
	OBJ_DIFF_MODE_HUGE
} OBJ_DIFF_MODE;


/**
 * Wrapper class around a memory block
 */
class MemoryBlock {
	private:
		/** Memory block */
		char * m_block;

		/** Length of the memory block */
		unsigned int m_length;

	public:

		/**
		 * Constructor for creating a memory block
		 */
		MemoryBlock();

		/**
		 * Destructor for destroying a memory block
		 */
		~MemoryBlock();

		/**
		 * Allocates the memory block.
		 *
		 * @param length   size of the memory block to be allocated
		 *
		 * @return pointer to the allocated memory block or NULL
		 */
		char * allocate( unsigned int length );


		/**
		 * Allocates or reallocates the memory block.
		 *
		 * @param length   size of the memory block to be allocated
		 *
		 * @return pointer to the allocated memory block or NULL
		 */
		char * reallocate( unsigned int length );


		/**
		 * Releases the allocated memory block
		 */
		void release();

		/**
		 * Chunks the memory block.
		 *
		 * @param newLength   new size for the memory block (should be less than or equals to the current length)
		 *
		 * @return pointer to the allocated memory block or NULL
		 */
		char * chunk( unsigned int newLength );

		/**
		 * Returns the allocated memory block.
		 *
		 * @return the allocated memory block
		 */
		inline const char * getBlock() const {
			return m_block;
		}

		/**
		 * Gets the length of the allocated memory block
		 *
		 * @return the length of the allocated memory block.
		 */
		inline unsigned int getLength() const {
			return m_length;
		}

};

/**
 * Memory manager for local objdiff: all images should be kept until get compared.
 */
class LocalObjDiffMemory : public CImageMemory {
	private:
		/** Container for memory blocks (index - memory block) */
		map< int, MemoryBlock * > m_blocks;

		/** Typedef for iterator */
		typedef map< int, MemoryBlock * >::iterator MBIterator;

	public:

		/**
		 * Constructor for creating a new LocalObjDiffMemory object.
		 */
		LocalObjDiffMemory();

		/**
		 * Destructor for destorying the LocalObjDiffMemory object.
		 */
		virtual ~LocalObjDiffMemory();

		/**
		 * Allocates (or reallocates) memory block for the given index.
		 *
		 * @param length   size of the memory block to be allocated
		 * @param index    index identifying the memory block
		 *
		 * @return pointer to the allocated memory block or NULL
		 */
		virtual char * allocate( unsigned int length, int index );

		/**
		 * Releases memory block for the given index.
		 *
		 * @param block    [not used]
		 * @param length   [not used]
		 * @param index    index identifying the memory block to be released
		 */
		virtual void release( char * block, unsigned int length, int index );

		/**
		 * Chunks memory block of the given index.
		 *
		 * @param block       [not used]
		 * @param oldlength   [not used]
		 * @param newLength   new size for the memory block (should be less than or equals to the current length)
		 * @param index       index identifying the memory block to be chunked
		 *
		 * @return pointer to the allocated memory block or NULL
		 */
		virtual char * chunk( char * block, unsigned int oldLength, unsigned int newLength, int index );

		/**
		 * Returns the collection of the memory blocks - to be used to compare processed pixel data
		 * 
		 * @return collection of memory blocks
		 */
		inline const map< int, MemoryBlock * > getBlocks() const {
			return m_blocks;
		}
};

/**
 * Memory manager for remote objdiff: a single memory block will be used and recycled for images will be written to file.
 */
class RemoteObjDiffMemory : public CImageMemory {
	private:
		char * m_block;
		unsigned int m_length;
		bool m_isInUse;
	public:
		RemoteObjDiffMemory();
		virtual ~RemoteObjDiffMemory();

		virtual char * allocate( unsigned int length, int index );
		virtual void release( char * block, unsigned int length, int index );
		virtual char * chunk( char * block, unsigned int oldLength, unsigned int newLength, int index );
};

class ObjDiffProcess : public CImageProcess {
	private:
		string m_outputPrefix;
		OBJ_DIFF_MODE m_mode;

	protected:
		virtual CRawImage * createRawImage();
		virtual CEncapsulatedDocument * createEncapsulatedDocument( DOCUMENT_TYPE type, Uint32 length );
		virtual OFBool postProcess( CExtractedData * data );

		virtual OFBool initProcess();
		virtual OFBool finalizeProcess();

	public:

		/**
		 * Constructor for local compare.
		 */
		ObjDiffProcess( DcmDataset * dataset, OFBool verbose );

		/**
		 * Constructor for remote compare.
		 */
		ObjDiffProcess( DcmDataset * dataset, OFBool verbose, OFBool dumpOnly, const char * outputPrefix );

		/**
		 * Constructor for comparing huge files
		 */
		ObjDiffProcess( DcmDataset * dataset, const OFString & outputPrefix, OFBool verbose );

		virtual ~ObjDiffProcess();

};

class ObjDiffRawImage : public CRawImage {
	protected:
		virtual void optimizeBitmap();
	public:
		ObjDiffRawImage( CImageProcess * imageProcess, int index );
		virtual ~ObjDiffRawImage();

		OFBool writeIcon( FILE * iconfile );
		OFBool writeWavelet( FILE * wavfile );
		OFBool writeRaw( FILE *rawfile );
};

/**
 * Maintains list of tags and tag patterns to be ignored during compare
 */
class IgnoreTags {
		set< DcmTagKey > m_ignoredkeys;
		set< string > m_regexps;

		bool matchesRegexp( const DcmTagKey & key ) const;

	public:

		/**
		 * Constructor for creating an empty ignore tags list
		 */
		IgnoreTags();

		/**
		 * Constructor for creating ignore tags list by reading tags and tag patterns from config file
		 * @param ignoreTagsCfg   ignore tags configuration file to read
		 */
		IgnoreTags( const char * ignoreTagsCfg );

		inline bool ignored( const DcmTagKey & key ) const {
			return m_ignoredkeys.count( key ) || matchesRegexp( key );
		}

		inline void addTag( const DcmTagKey & key ) {
			m_ignoredkeys.insert( key );
		}

		inline void ignoreGroupLength() {
			m_regexps.insert( "....0000" );
		}

		/**
		 * Add an eRAD private tag (in the f215 group) to the ignore list.
		 */
		void ignorePB( Uint16 pbElem );
};

class DumpImage {
		unsigned char *data;
		int bpp;
		int width;
		int height;
	public:
		DumpImage() :
			data( NULL ), bpp( 0 ), width( 0 ), height( 0 ) {
		}
		DumpImage( string name );
		~DumpImage();

		bool loadImage( string name );
		bool operator==( const DumpImage& other ) const;
};

/**
 * Pixel Data Compare modes
 */
typedef enum {
	/** Bitwise compare - Pixel Data expected to be in the same TS */
	COMPARE_BITWISE,
	/** Regular compare using image preprocessing */
	COMPARE_REGULAR
} OC_MODE;


/**
 * Wrapper around a huge object. Used when comparing huge objects.
 *
 * Comparing huge objects in the regular way runs out of memory on 32 bit models.
 * Image frames are extracted to temporary directories and compared one by one
 * instead of extracting both objects to the memory.
 *
 * Usage:
 *
 * HugeObject hugeObject( objectFile [, pbrFile] );
 * if ( !extractTo( tempDir ) ) {
 *    // error handling
 *    ...
 * }
 *
 * // use getters
 *
 * hugeObject.get...()
 *
 * // or access frames
 *
 * MemoryBlock * mb = hugeObject.getMemoryBlock();
 * for ( i = ... ) {
 *    if ( ! hugeObject.readFrame( i ) ) {
 *       // error handling
 *       ...
 *    }
 *
 *    mb.getLength(); // the length of the [i]th frame in bytes
 *    mb.getBlock(); // the pixel data of the [i]th frame
 * }
 *
 * Note:
 *  Reading a frame puts data to the same memory block. The data block is reallocated if needed.
 */
class HugeObject {
	private:
		/** DICOM file format to read the object to */
		DcmFileFormat dcmff;
		/** Image file processor to convert and extract image frames */
		ObjDiffProcess * process;
		/** The original Transfer Syntax of the object */
		E_TransferSyntax origTS;

		/** Path to the huge object to process */
		const char * objFileName;
		/** Apply soft edit changes from this file to the object (if provided) */
		const char * pbrFileName;

		/** Temporary working directory to save the image frames to */
		OFString workDir;

		/** Memory block to read the image frames one by one to for the comparison */
		MemoryBlock mb;

	public:
		HugeObject( const char *& objFile, const char *& pbrFile );
		~HugeObject();

		bool extractTo( const OFString & dir );

		inline DcmDataset * getDataset() {
			return dcmff.getDataset();
		}

		inline E_TransferSyntax & getOrigTS() {
			return origTS;
		}

		inline ObjDiffProcess & getProcessInfo() {
			return *process;
		}

		inline MemoryBlock * getMemoryBlock() {
			return &mb;
		}

		bool readFrame( Uint16 frameNdx );
};



/**
 * Comparing DICOM objects
 */
class ObjectComparator {
	private:
		/** List of tags to be ignored during compare */
		IgnoreTags & m_ignoreTags;

		/** Image compression codecs */
		CodecRegistry * m_codecs;

		/** Verbose mode: print verbose messages */
		bool m_verbose;

		/**
		 * Compare binary (OB/OW) elements to each other. The value of the two elements will be binary compared.
		 * The values of the elements are the same if both element has the same value length, the same VR
		 * and the values binary equal.
		 *
		 * @param obow1   One binary element to compare
		 * @param obow2   Other binary element to compare
		 *
		 * @return true if and only if the values of the elements equal.
		 */
		bool compareOBOW( DcmOtherByteOtherWord * obow1, DcmOtherByteOtherWord * obow2 );

		/**
		 * Compares elements to each other. Binary (OB/OW) elements will be compared binary.
		 * In case of non binary elements the string representations of the elements will be compared.
		 *
		 * @param elem1   One element to compare
		 * @param elem2   Other element to compare
		 *
		 * @return true if and only if the values of the elements equal.
		 */
		bool compareElements( DcmElement * elem1, DcmElement * elem2 );

		/**
		 * Compares pixel sequences to each other.
		 * The pixel sequences are the same if both sequences has the same number of pixel items, and the pixel items binary equal respectively.

		 * @param pixelSeq1  One pixel sequence to compare
		 * @param pixelSeq2  Other pixel sequence to compare
		 *
		 * @return true if and only if the pixel sequences equal.
		 */
		bool comparePixelSequence( DcmPixelSequence * pixelSeq1, DcmPixelSequence * pixelSeq2 );

		/**
		 * Binary compares to pixel data element to each other.
		 * The pixel data elements equal if
		 *   - either both pixel data elements consist of pixel sequences which equal,
		 *   - or both are OB/OW and binary equals
		 *
		 * @param pixelData1   One pixel data element to compare
		 * @param pixelData2   Other pixel data element to compare
		 *
		 * @return true if and only if the pixel data elements equal
		 */
		bool comparePixelDataElement( DcmPixelData * pixelData1, DcmPixelData * pixelData2 );

		/**
		 * Compares datasets of DICOM objects to each other.
		 * The datasets equal if all elements in the dataset (except ones listed in Ignore Tags) equal.
		 * Pixel data will be binary compared if comparePixelData is true, and will not be compared if
		 * comparePixelData is false. In the latter case pixel data need to be compared by the caller.
		 *
		 * @param ds1                Dataset of one object to compare
		 * @param ds2                Dataset of other object to compare
		 * @param comparePixelData  whether Pixel Data need to be binary compared or not
		 *
		 * @return true if and only if the pixel data elements equal
		 */
		bool compareDataset( DcmDataset * ds1, DcmDataset * ds2, bool comparePixelData );

		/**
		 * Compares preprocessed Pixel Data frames to each other.
		 *
		 * @param mb1         Memory block holding one pixel data frame to compare
		 * @param mb2         Memory block holding other pixel data frame to compare
		 * @param isBitWise   If true pixel data frames will be compared lossless (bitwise), otherwise lossy
		 * @param isWord      True indicates frame contains word (OW) values, otherwise byte (OB) values
		 * @return true if and only if the pixel data frames equal
		 */
		bool compareMemoryBlocks( MemoryBlock * mb1, MemoryBlock * mb2, bool isBitWise, bool isWord );

		/**
		 * Checks if Transfer Syntax is lossless or not
		 * LossLess:
		 *    EXS_LittleEndianImplicit
		 *    EXS_BigEndianImplicit
		 *    EXS_LittleEndianExplicit
		 *    EXS_BigEndianExplicit
		 *    EXS_JPEGProcess14TransferSyntax
		 *    EXS_JPEGProcess14SV1TransferSyntax
		 *    EXS_RLELossless
		 *    EXS_JPEGLSLossless
		 *    EXS_JPEG2000LosslessOnly
		 *    EXS_JPEG2000MulticomponentLosslessOnly
		 *
		 * @param ts Transfer Syntax to check
		 *
		 * @return true if and only if Transfer Syntax is lossless
		 */
		bool isLossLess( E_TransferSyntax ts );

		/**
		 * Register image compression codecs.
		 */
		inline void checkAndRegisterCodecs() {
			if ( m_codecs == NULL ) {
				m_codecs = new CodecRegistry();
			}
		}

//		bool extractTo( DcmFileFormat & dcmff, const char * objFile, const char * pbrFile, const OFString & dir );

	public:
		/**
		 * Constructor for creating a new Object Comparator object.
		 *
		 * @param ignoreTags  Tags to be ignored during comparison
		 */
		ObjectComparator( IgnoreTags & ignoreTags );

		/**
		 * Destructor for destroying the Object Comparator object.
		 */
		~ObjectComparator();

		/**
		 * Sets verbose mode to turn on/off printing verbose messages to stnadard output.
		 * Verbose mode is useful for debugging.
		 *
		 * @param verbose  new value for verbose mode (true for on, false for off)
		 */
		inline void setVerboseMode( bool verbose ) {
			m_verbose = verbose;
		}

		/**
		 * Compares image (Pixel Data) dumps created by one of the dumpObject methods to each other.
		 * Image dumps equal will be compared using lossy comparison.
		 *
		 * @param dump1  One image dump to compare
		 * @param dump2  Other image dump to compare
		 *
		 * @return true if and only if the image dumps equal
		 */
		bool compareDumps( const char * dump1, const char * dump2 );

		/**
		 * Will dump a DICOM object to the standard output (excluding elements listed in Ignore Tags).
		 * Pixel data will be written to outputFile.
		 *
		 * @param dataset     Dataset of the DICOM object to be dumped
		 * @param outputFile  Filename for the Pixel Data to be printed to
		 *
		 * @return true if and only if the DICOM object and pixel data was dumped successfully
		 */
		bool dumpObject( DcmDataset * dataset, const char * outputFile );

		/**
		 * Will dump a DICOM object to the standard output (excluding elements listed in Ignore Tags).
		 * If PbR file is provided, the PbR will also be loaded and soft edit changes will be applied to the
		 * object to be dumped before dumping.
		 * Pixel data will be written to outputFile.
		 *
		 * @param objFile      Path to the DICOM object to be dumped
		 * @param pbrFile      Path to PbR file to apply soft edit changes to the DICOM object (may be NULL)
		 * @param outputFile   Filename for the Pixel Data to be printed to
		 *
		 * @return true if and only if the DICOM object and pixel data was dumped successfully
		 */
		bool dumpObject( const char * objFile, const char * pbrFile, const char * outputFile );

		/**
		 * Compares the datasets of DICOM objects to each other. Elements listed in Ignore Tags will be excluded.
		 * Based on comparison mode, Pixel Data will be compared binary (Pixel Data expected to be in the same Transfer Syntax)
		 * or the regular way (comparing preprocessed Pixel Data using Lossless comparison of both objects are in
		 * Lossless transfer syntax, and lossy comparison if any of the transfer syntaxes is lossy)
		 *
		 * @param ds1   Dataset of one object to compare
		 * @param ds2   Dataset of other object to compare
		 * @param mode  Comparison mode
		 *
		 * @return true if and only if the datasets equal
		 */
		bool compareObjects( DcmDataset * ds1, DcmDataset * ds2, OC_MODE mode );

		/**
		 * Compares DICOM objects to each other. Elements listed in Ignore Tags will be excluded.
		 * If a PbR file is provided soft edit changes will be applied to the corrsponding DICOM object before the comparison.
		 * Based on comparison mode, Pixel Data will be compared binary (Pixel Data expected to be in the same Transfer Syntax)
		 * or the regular way (comparing preprocessed Pixel Data using Lossless comparison of both objects are in
		 * Lossless transfer syntax, and lossy comparison if any of the transfer syntaxes is lossy)
		 *
		 * @param objFile1    Path to one DICOM object to be compared
		 * @param pbrFile1    Path to PbR file to apply soft edit changes to objFile1 (may be NULL)
		 * @param objFile2    Path to other DICOM object to be compared
		 * @param pbrFile2    Path to PbR file to apply soft edit changes to objFile2 (may be NULL)
		 * @param mode        Comparison mode
		 *
		 * @return true if and only if the DICOM objects (having soft edit changes applied) equal.
		 */
		bool compareObjects( const char * objFile1, const char * pbrFile1, const char * objFile2, const char * pbrFile2,
				OC_MODE mode );


		/**
		 * Compares DICOM objects to each other. Elements listed in Ignore Tags will be excluded.
		 * If a PbR file is provided soft edit changes will be applied to the corresponding DICOM object before the comparison.
		 * Pixel Data will be compared the regular way (comparing preprocessed Pixel Data using Lossless comparison of both objects are in
		 * Lossless transfer syntax, and lossy comparison if any of the transfer syntaxes is lossy)
		 *
		 * Huge objects might not fit to the memory, the dataset and the images extracted from the objects will be
		 * saved to temporary files and will be compared from there
		 *
		 * @param obj1        Path to one of the huge objects and the corresponding PbR (if any)
		 * @param obj2        Path to the other huge object and the corresponding PbR (if any)
		 *
		 * @return true if and only if the huge objects (having soft edit changes applied) equal.
		 */
		bool compareHugeObjects( HugeObject & obj1, HugeObject & obj2 );


};


/**
 * Loads a DICOM object to a DcmFileFormat. If PbR file is also provided (not NULL), it will also be loaded and
 * soft edit changes be applied to the to the loaded DICOM object.
 *
 * @param fileformat   Output parameter: content of the DICOM object will be loaded to it
 * @param fileName     Path to the DICOM object to be loaded
 * @param pbrName      Path to PbR file to apply soft edit changes to the DICOM object loaded (may be NULL)
 *
 * @return true if and only if the DICOM object was successfully loaded and soft edit changes (if any) applied to the loaded object.
 */
bool loadObjectAndPbR( DcmFileFormat & fileformat, const char * fileName, const char * pbrName );




template< class A, int n >
struct Pow {
		static inline A pow( A base ) {
			return base * Pow< A, n - 1 >::pow( base );
		}
};
template< class A >
struct Pow< A, 1 > {
		static inline A pow( A base ) {
			return base;
		}
};

#define SAME_IMAGE_THRESHOLD 40
template< class A >
double compareImageData( A *data1, A *data2, unsigned int l ) {
	long double error = 0;
	double max = 0;

	for ( unsigned int i = 0; i < l; data1++, data2++, i++ ) {
		const int &d1 = *data1;
		const int &d2 = *data2;

		max = max < d1 ? d1 : max;
		error += Pow< long double, MSE_POW >::pow( abs( d1 - d2 ) );
	}

	error = error / (long double) l / Pow< long double, MSE_POW >::pow( max );

	//	cout << " Max :" << max << endl;
	//	cout << " error ::" << error << endl;

	double goodness = -10.0 * logl( error ) / logl( 10.0 );

	//	cout << "Goodness :" << goodness << endl;

	return goodness;
}

#endif

