/**
 * Component: utils
 *
 * Description:
 *    Helper methods to start tasks
 *
 *  TODO: on long term these methods should be moved to component taskd
 *
 * Last Updated     : $Author$
 * Update Date Time : $Date$
 * RCS File Name    : $RCSfile$
 * RCS Revision     : $Revision$
 * Locked by        : $Locker$
 * Symbolic name    : $Name$
 * State            : $State$
 */

#ifndef TASKHELPER_H_
#define TASKHELPER_H_

#include "dcmtk/ofstd/ofstd.h"

/**
 * Reindex modes
 */
typedef enum {
	/** Regular store - no reindex */
	STORESCP_REG_OPT_NONE,
	/** Reindex to the database only */
	STORESCP_REG_OPT_REINDEX,
	/** Reindex and autoforward, ...*/
	STORESCP_REG_OPT_FULL,
	/** Called from dcfixall (implies full)*/
	STORESCP_REG_OPT_DCFIXALL
} STORESCP_REG_OPT;

bool runOrScheduleJavaTask( const char * priority, const char * cmd, const char * delta );

class StorescpRegStarter {
	private:

		/** Instance UID of the study the object to be stored for */
		OFString m_study;
		/** File name of the object */
		OFString m_file;
		/** Reindex mode */
		STORESCP_REG_OPT m_reindexMode;
		/** Path to the object file to be stored */
		OFString m_tempFile;
		/** true indicates to try to register the directly */
		bool m_isDirect;
		/** Is double check after moving to repository - used when migrating from classic archive to repository */
		bool m_isDoubleCheck;
		/** Indicates whether registration tasks were started from obsolete tool */
		bool m_isStartedFromObsolete;

		/** ID of the sender device */
		OFString m_senderDevice;
		/** IP address of the sender device */
		OFString m_senderIP;
		/** AE title of the sender device */
		OFString m_senderAE;

		// Taskd related properties - may be moved to a TaskStarter class


		/** Delta priority */
		OFString m_delta;
		/** Job ID - Identifies the job the task belongs to */
		OFString m_jobId;
		/** Sub Job ID - Identifies the sub job the task belongs to */
		OFString m_subJobId;
		/** Child ID  - Identifies the child task within sub job */
		OFString m_childId;
		/** Job Information */
		OFString m_jobInfo;
		/** Relative Priority */
		int m_relativePriority;

	public:

		/**
		 * Constructor creating a new StorescpRegStarter object
		 *
		 * @param study         Instance UID of the study the object to be stored for
		 * @param file          File name of the object
		 * @param reindexMode   Reindex mode
		 * @param tempFile      Path to the object file to be stored
		 */
		StorescpRegStarter( const OFString & study, const OFString & file, STORESCP_REG_OPT reindexMode, const OFString & tempFile );

		/**
		 * Destructor destroying this StorescpRegStarter object
		 */
		~StorescpRegStarter();

		/**
		 * Sets the ID of the device the object was received from.
		 *
		 * @param senderDevice  ID of the device the object was received from
		 */
		inline void setSenderDevice( const char * senderDevice ) {
			m_senderDevice = senderDevice == NULL ? "" : senderDevice;
		}


		/**
		 * Sets the IP address of the device the object was received from.
		 * To be used only in promiscuous mode for unregistered devices,
		 * or in case of objects received from archive
		 *
		 * @param senderIP  IP address of the device the object was received from
		 */
		inline void setSenderIP( const char * senderIP ) {
			m_senderIP = senderIP == NULL ? "" : senderIP;
		}

		/**
		 * Sets the AE title of the device the object was received from.
		 * To be used only in promiscuous mode for unregistered devices.
		 *
		 * @param senderAE  AE title of the device the object was received from
		 */
		inline void setSenderAE( const char * senderAE ) {
			m_senderAE = senderAE == NULL ? "" : senderAE;
		}

		/**
		 * Sets whether the object should be registered directly or not (placing to the queue)
		 * @param isDirect  true indicates to register the object directly
		 */
		inline void setIsDirect( bool isDirect ) {
			m_isDirect = isDirect;
		}

		/**
		 * Sets whether the object should be double-checked after moving to repository
		 * (migration from classic archive to repository archive)
		 * @param isDoubleCheck  true indicates to double-check object when moving to repository
		 */
		inline void setIsDoubleCheck( bool isDoubleCheck ) {
			m_isDoubleCheck = isDoubleCheck;
		}

		/**
		 * Sets whether the task is started from obsolete tool
		 * ( if true then imagemed/dcprocessing/StudyRemover.postObsolete() Java method should not be invoked )
		 * @param isStartedFromObsolete true indicates the task is started from obsolete tool
		 */
		inline void setIsStartedFromObsolete ( bool isStartedFromObsolete ){
			m_isStartedFromObsolete = isStartedFromObsolete;
		}

		/**
		 * Sets Delta priority to be used for creating the task
		 *
		 * @param delta  Delta priority to be used for creating the task
		 */
		inline void setDelta( const char * delta ) {
			m_delta = delta == NULL ? "" : delta;
		}

		/**
		 * Sets the job ID the task to be added to
		 *
		 * @param jobId  ID of the job the task to be added to
		 */
		inline void setJobId( const char * jobId ) {
			m_jobId = jobId == NULL ? "" : jobId;
		}

		/**
		 * Sets the subjob ID the task to be added to
		 *
		 * @param subJobId  ID of the subjob the task to be added to
		 */
		inline void setSubJobId( const char * subJobId ) {
			m_subJobId = subJobId == NULL ? "" : subJobId;
		}

		/**
		 * Sets child ID for the task
		 *
		 * @param childId  child ID to be used when constructing task name
		 */
		inline void setChildId( const char * childId ) {
			m_childId = childId == NULL ? "" : childId;
		}

		/**
		 * Sets job information
		 *
		 * @param jobInfo  job information
		 */
		inline void setJobInfo( const char * jobInfo ) {
			m_jobInfo = jobInfo == NULL ? "" : jobInfo;
		}

		/**
		 * Sets relative priority to be used
		 *
		 * @param relativePriority   Relative priority
		 */
		inline void setRelativePriority( int relativePriority ) {
			m_relativePriority = relativePriority;
		}

		/**
		 * Place the object to data/tmp, create and schedule the StorescpReg task
		 *
		 * @return  true if and only if the task was successfully created
		 */
		bool startTask();
};

#endif /* TASKHELPER_H_ */
